from fastmcp import FastMCP
from fastmcp.server.dependencies import get_context
from ..agent import select_tool
from pydantic import Field

auto_mcp = FastMCP("SmartMCP-select-Server")


@auto_mcp.tool(tags={"auto"})
def search_tool(
    task: str = Field(
        description="The tasks or questions that needs to be solved using available tools"
    ),
):
    """search the tools and get tool parameters that can be used to solve the  user's tasks or questions"""
    ctx = get_context()
    fastmcp = ctx.fastmcp
    if hasattr(fastmcp._tool_manager, "_all_tools"):
        all_tools = fastmcp._tool_manager._all_tools
    else:
        all_tools = fastmcp._tool_manager._tools
    auto_tools = fastmcp._tool_manager._tools
    fastmcp._tool_manager._tools = all_tools
    query = f"<task>{task}</task>\n"
    for name in all_tools:
        tool = all_tools[name]
        query += f"<Tool>\n<name>{name}</name>\n<description>{tool.description}</description>\n</Tool>\n"
    results = select_tool(query)
    tool_list = []
    for tool in results.tools:
        tool = tool.model_dump()
        tool["parameters"] = all_tools[tool["name"]].parameters
        tool_list.append(tool)
    fastmcp._tool_manager._tools = auto_tools
    return tool_list


@auto_mcp.tool(tags={"auto"})
async def run_tool(
    name: str = Field(description="The name of the tool to run"),
    parameter: dict = Field(description="The parameters to pass to the tool"),
):
    """run the tool with the given name and parameters. Only start call the tool when last tool is finished."""
    ctx = get_context()
    fastmcp = ctx.fastmcp
    all_tools = fastmcp._tool_manager._all_tools
    auto_tools = fastmcp._tool_manager._tools
    fastmcp._tool_manager._tools = all_tools

    try:
        result = await fastmcp._tool_manager.call_tool(name, parameter)
    except Exception as e:
        fastmcp._tool_manager._tools = auto_tools
        result = {"error": str(e)}

    return result



from pydantic import Field, field_validator, ValidationInfo, BaseModel
from typing import Optional, Union, List, Dict, Any, Literal


class TSNEParam(BaseModel):
    """Input schema for the t-SNE dimensionality reduction tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    n_pcs: Optional[int] = Field(
        default=None,
        description="Number of PCs to use. If None, automatically determined.",
        ge=0,
    )
    use_rep: Optional[str] = Field(
        default=None, description="Key for .obsm to use as representation."
    )
    perplexity: Optional[Union[float, int]] = Field(
        default=30,
        description="Related to number of nearest neighbors, typically between 5-50.",
        gt=0,
    )
    early_exaggeration: Optional[Union[float, int]] = Field(
        default=12,
        description="Controls space between natural clusters in embedded space.",
        gt=0,
    )
    learning_rate: Optional[Union[float, int]] = Field(
        default=1000,
        description="Learning rate for optimization, typically between 100-1000.",
        gt=0,
    )
    use_fast_tsne: Optional[bool] = Field(
        default=False, description="Whether to use Multicore-tSNE implementation."
    )
    n_jobs: Optional[int] = Field(
        default=None, description="Number of jobs for parallel computation.", gt=0
    )
    metric: Optional[str] = Field(
        default="euclidean", description="Distance metric to use."
    )

    @field_validator(
        "n_pcs", "perplexity", "early_exaggeration", "learning_rate", "n_jobs"
    )
    def validate_positive_numbers(
        cls, v: Optional[Union[int, float]]
    ) -> Optional[Union[int, float]]:
        """Validate positive numbers where applicable"""
        if v is not None and v <= 0:
            raise ValueError("must be a positive number")
        return v

    @field_validator("metric")
    def validate_metric(cls, v: str) -> str:
        """Validate distance metric is supported"""
        valid_metrics = ["euclidean", "cosine", "manhattan", "l1", "l2"]
        if v.lower() not in valid_metrics:
            raise ValueError(f"metric must be one of {valid_metrics}")
        return v.lower()


class UMAPParam(BaseModel):
    """Input schema for the UMAP dimensionality reduction tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    min_dist: Optional[float] = Field(
        default=0.5, description="Minimum distance between embedded points.", gt=0
    )
    spread: Optional[float] = Field(
        default=1.0, description="Scale of embedded points.", gt=0
    )
    n_components: Optional[int] = Field(
        default=2, description="Number of dimensions of the embedding.", gt=0
    )
    maxiter: Optional[int] = Field(
        default=None,
        description="Number of iterations (epochs) of the optimization.",
        gt=0,
    )
    alpha: Optional[float] = Field(
        default=1.0,
        description="Initial learning rate for the embedding optimization.",
        gt=0,
    )
    gamma: Optional[float] = Field(
        default=1.0, description="Weighting applied to negative samples.", gt=0
    )
    negative_sample_rate: Optional[int] = Field(
        default=5, description="Number of negative samples per positive sample.", gt=0
    )
    init_pos: Optional[str] = Field(
        default="spectral",
        description="How to initialize the low dimensional embedding.",
    )
    random_state: Optional[int] = Field(
        default=0, description="Random seed for reproducibility."
    )
    a: Optional[float] = Field(
        default=None, description="Parameter controlling the embedding.", gt=0
    )
    b: Optional[float] = Field(
        default=None, description="Parameter controlling the embedding.", gt=0
    )
    method: Optional[str] = Field(
        default="umap", description="Implementation to use ('umap' or 'rapids')."
    )
    neighbors_key: Optional[str] = Field(
        default=None, description="Key for neighbors settings in .uns."
    )

    @field_validator(
        "min_dist",
        "spread",
        "n_components",
        "maxiter",
        "alpha",
        "gamma",
        "negative_sample_rate",
        "a",
        "b",
    )
    def validate_positive_numbers(
        cls, v: Optional[Union[int, float]]
    ) -> Optional[Union[int, float]]:
        """Validate positive numbers where applicable"""
        if v is not None and v <= 0:
            raise ValueError("must be a positive number")
        return v

    @field_validator("method")
    def validate_method(cls, v: str) -> str:
        """Validate implementation method is supported"""
        if v.lower() not in ["umap", "rapids"]:
            raise ValueError("method must be either 'umap' or 'rapids'")
        return v.lower()


class DrawGraphParam(BaseModel):
    """Input schema for the force-directed graph drawing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    layout: str = Field(
        default="fa",
        description="Graph layout algorithm ('fa', 'fr', 'kk', etc.)",
    )
    init_pos: Optional[Union[str, bool]] = Field(
        default=None,
        description="Initial position for nodes ('paga'/True, False, or .obsm key)",
    )
    root: Optional[int] = Field(
        default=None, description="Root node for tree layouts", ge=0
    )
    random_state: int = Field(default=0, description="Random seed for reproducibility")
    n_jobs: Optional[int] = Field(
        default=None, description="Number of jobs for parallel computation", gt=0
    )
    key_added_ext: Optional[str] = Field(
        default=None, description="Suffix for storing results in .obsm"
    )
    neighbors_key: Optional[str] = Field(
        default=None, description="Key for neighbors settings in .uns"
    )
    obsp: Optional[str] = Field(
        default=None, description="Key for adjacency matrix in .obsp"
    )

    @field_validator("layout")
    def validate_layout(cls, v: str) -> str:
        """Validate layout is supported"""
        valid_layouts = ["fa", "fr", "grid_fr", "kk", "lgl", "drl", "rt"]
        if v.lower() not in valid_layouts:
            raise ValueError(f"layout must be one of {valid_layouts}")
        return v.lower()

    @field_validator("root", "n_jobs")
    def validate_positive_integers(cls, v: Optional[int]) -> Optional[int]:
        """Validate positive integers where applicable"""
        if v is not None and v <= 0:
            raise ValueError("must be a positive integer")
        return v


class DiffMapParam(BaseModel):
    """Input schema for the Diffusion Maps dimensionality reduction tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    n_comps: int = Field(
        default=15, description="The number of dimensions of the representation.", gt=0
    )
    neighbors_key: Optional[str] = Field(
        default=None,
        description=(
            "If not specified, diffmap looks .uns['neighbors'] for neighbors settings "
            "and .obsp['connectivities'], .obsp['distances'] for connectivities and "
            "distances respectively. If specified, diffmap looks .uns[neighbors_key] for "
            "neighbors settings and uses the corresponding connectivities and distances."
        ),
    )
    random_state: int = Field(default=0, description="Random seed for reproducibility.")

    @field_validator("n_comps")
    def validate_positive_integers(cls, v: int) -> int:
        """Validate positive integers"""
        if v <= 0:
            raise ValueError("n_comps must be a positive integer")
        return v


class EmbeddingDensityParam(BaseModel):
    """Input schema for the embedding density calculation tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    basis: str = Field(
        default="umap",
        description="The embedding over which the density will be calculated. This embedded representation should be found in `adata.obsm['X_[basis]']`.",
    )
    groupby: Optional[str] = Field(
        default=None,
        description="Key for categorical observation/cell annotation for which densities are calculated per category.",
    )
    key_added: Optional[str] = Field(
        default=None,
        description="Name of the `.obs` covariate that will be added with the density estimates.",
    )
    components: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="The embedding dimensions over which the density should be calculated. This is limited to two components.",
    )

    @field_validator("components")
    def validate_components(
        cls, v: Optional[Union[str, List[str]]]
    ) -> Optional[Union[str, List[str]]]:
        """Validate that components are limited to two dimensions"""
        if v is not None and isinstance(v, list) and len(v) > 2:
            raise ValueError("components is limited to two dimensions")
        return v


class LeidenParam(BaseModel):
    """Input schema for the Leiden clustering algorithm."""

    adata: str = Field(..., description="The AnnData object variable name.")
    resolution: Optional[float] = Field(
        default=1.0,
        description="A parameter value controlling the coarseness of the clustering. Higher values lead to more clusters.",
    )
    key_added: Optional[str] = Field(
        default="leiden",
        description="`adata.obs` key under which to add the cluster labels.",
    )
    directed: Optional[bool] = Field(
        default=None,
        description="Whether to treat the graph as directed or undirected.",
    )
    use_weights: Optional[bool] = Field(
        default=True,
        description="If `True`, edge weights from the graph are used in the computation (placing more emphasis on stronger edges).",
    )
    n_iterations: Optional[int] = Field(
        default=-1,
        description="How many iterations of the Leiden clustering algorithm to perform. -1 runs until optimal clustering.",
    )
    neighbors_key: Optional[str] = Field(
        default=None,
        description="Use neighbors connectivities as adjacency. If specified, leiden looks .obsp[.uns[neighbors_key]['connectivities_key']] for connectivities.",
    )
    obsp: Optional[str] = Field(
        default=None,
        description="Use .obsp[obsp] as adjacency. You can't specify both `obsp` and `neighbors_key` at the same time.",
    )
    flavor: Optional[Literal["leidenalg", "igraph"]] = Field(
        default="igraph", description="Which package's implementation to use."
    )

    @field_validator("resolution")
    def validate_resolution(cls, v: float) -> float:
        """Validate resolution is positive"""
        if v <= 0:
            raise ValueError("resolution must be a positive number")
        return v

    @field_validator("obsp", "neighbors_key")
    def validate_graph_source(
        cls, v: Optional[str], info: ValidationInfo
    ) -> Optional[str]:
        """Validate that obsp and neighbors_key are not both specified"""
        values = info.data
        if v is not None and "obsp" in values and "neighbors_key" in values:
            if values["obsp"] is not None and values["neighbors_key"] is not None:
                raise ValueError("Cannot specify both obsp and neighbors_key")
        return v

    @field_validator("flavor")
    def validate_flavor(cls, v: str) -> str:
        """Validate flavor is supported"""
        if v not in ["leidenalg", "igraph"]:
            raise ValueError("flavor must be either 'leidenalg' or 'igraph'")
        return v


class LouvainParam(BaseModel):
    """Input schema for the Louvain clustering algorithm."""

    adata: str = Field(..., description="The AnnData object variable name.")
    resolution: Optional[float] = Field(
        default=None,
        description="For the default flavor ('vtraag') or for 'RAPIDS', you can provide a resolution (higher resolution means finding more and smaller clusters), which defaults to 1.0.",
    )
    random_state: int = Field(
        default=0, description="Change the initialization of the optimization."
    )
    key_added: str = Field(
        default="louvain", description="Key under which to add the cluster labels."
    )
    flavor: Literal["vtraag", "igraph", "rapids"] = Field(
        default="vtraag",
        description="Package for computing the clustering: 'vtraag' (default, more powerful), 'igraph' (built-in method), or 'rapids' (GPU accelerated).",
    )
    directed: bool = Field(
        default=True, description="Interpret the adjacency matrix as directed graph."
    )
    use_weights: bool = Field(default=False, description="Use weights from knn graph.")
    partition_kwargs: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Key word arguments to pass to partitioning, if 'vtraag' method is being used.",
    )
    neighbors_key: Optional[str] = Field(
        default=None,
        description="Use neighbors connectivities as adjacency. If specified, louvain looks .obsp[.uns[neighbors_key]['connectivities_key']] for connectivities.",
    )
    obsp: Optional[str] = Field(
        default=None,
        description="Use .obsp[obsp] as adjacency. You can't specify both `obsp` and `neighbors_key` at the same time.",
    )

    @field_validator("resolution")
    def validate_resolution(cls, v: Optional[float]) -> Optional[float]:
        """Validate resolution is positive if provided"""
        if v is not None and v <= 0:
            raise ValueError("resolution must be a positive number")
        return v

    @field_validator("obsp", "neighbors_key")
    def validate_graph_source(
        cls, v: Optional[str], info: ValidationInfo
    ) -> Optional[str]:
        """Validate that obsp and neighbors_key are not both specified"""
        values = info.data
        if v is not None and "obsp" in values and "neighbors_key" in values:
            if values["obsp"] is not None and values["neighbors_key"] is not None:
                raise ValueError("Cannot specify both obsp and neighbors_key")
        return v

    @field_validator("flavor")
    def validate_flavor(cls, v: str) -> str:
        """Validate flavor is supported"""
        if v not in ["vtraag", "igraph", "rapids"]:
            raise ValueError("flavor must be one of 'vtraag', 'igraph', or 'rapids'")
        return v


class DendrogramParam(BaseModel):
    """Input schema for the hierarchical clustering dendrogram tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    groupby: str = Field(
        ...,  # Required field
        description="The categorical observation annotation to use for grouping.",
    )
    n_pcs: Optional[int] = Field(
        default=None,
        description="Use this many PCs. If n_pcs==0 use .X if use_rep is None.",
        ge=0,
    )
    use_rep: Optional[str] = Field(
        default=None,
        description="Use the indicated representation. 'X' or any key for .obsm is valid.",
    )
    var_names: Optional[List[str]] = Field(
        default=None,
        description="List of var_names to use for computing the hierarchical clustering. If provided, use_rep and n_pcs are ignored.",
    )
    use_raw: Optional[bool] = Field(
        default=None,
        description="Only when var_names is not None. Use raw attribute of adata if present.",
    )
    cor_method: str = Field(
        default="pearson",
        description="Correlation method to use: 'pearson', 'kendall', or 'spearman'.",
    )
    linkage_method: str = Field(
        default="complete",
        description="Linkage method to use for hierarchical clustering.",
    )
    optimal_ordering: bool = Field(
        default=False,
        description="Reorders the linkage matrix so that the distance between successive leaves is minimal.",
    )
    key_added: Optional[str] = Field(
        default=None,
        description="By default, the dendrogram information is added to .uns[f'dendrogram_{groupby}'].",
    )

    @field_validator("cor_method")
    def validate_cor_method(cls, v: str) -> str:
        """Validate correlation method is supported"""
        valid_methods = ["pearson", "kendall", "spearman"]
        if v.lower() not in valid_methods:
            raise ValueError(f"cor_method must be one of {valid_methods}")
        return v.lower()

    @field_validator("linkage_method")
    def validate_linkage_method(cls, v: str) -> str:
        """Validate linkage method is supported"""
        valid_methods = [
            "single",
            "complete",
            "average",
            "weighted",
            "centroid",
            "median",
            "ward",
        ]
        if v.lower() not in valid_methods:
            raise ValueError(f"linkage_method must be one of {valid_methods}")
        return v.lower()

    @field_validator("n_pcs")
    def validate_n_pcs(cls, v: Optional[int]) -> Optional[int]:
        """Validate n_pcs is non-negative"""
        if v is not None and v < 0:
            raise ValueError("n_pcs must be a non-negative integer")
        return v


class DPTParam(BaseModel):
    """Input schema for the Diffusion Pseudotime (DPT) tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    n_dcs: int = Field(
        default=10, description="The number of diffusion components to use.", gt=0
    )
    n_branchings: int = Field(
        default=0, description="Number of branchings to detect.", ge=0
    )
    min_group_size: float = Field(
        default=0.01,
        description="During recursive splitting of branches, do not consider groups that contain less than min_group_size data points. If a float, refers to a fraction of the total number of data points.",
        gt=0,
        le=1.0,
    )
    allow_kendall_tau_shift: bool = Field(
        default=True,
        description="If a very small branch is detected upon splitting, shift away from maximum correlation in Kendall tau criterion to stabilize the splitting.",
    )
    neighbors_key: Optional[str] = Field(
        default=None,
        description="If specified, dpt looks .uns[neighbors_key] for neighbors settings and uses the corresponding connectivities and distances.",
    )

    @field_validator("n_dcs")
    def validate_n_dcs(cls, v: int) -> int:
        """Validate n_dcs is positive"""
        if v <= 0:
            raise ValueError("n_dcs must be a positive integer")
        return v

    @field_validator("n_branchings")
    def validate_n_branchings(cls, v: int) -> int:
        """Validate n_branchings is non-negative"""
        if v < 0:
            raise ValueError("n_branchings must be a non-negative integer")
        return v

    @field_validator("min_group_size")
    def validate_min_group_size(cls, v: float) -> float:
        """Validate min_group_size is between 0 and 1"""
        if v <= 0 or v > 1:
            raise ValueError("min_group_size must be between 0 and 1")
        return v


class PAGAParam(BaseModel):
    """Input schema for the Partition-based Graph Abstraction (PAGA) tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    groups: Optional[str] = Field(
        default=None,
        description="Key for categorical in adata.obs. You can pass your predefined groups by choosing any categorical annotation of observations. Default: The first present key of 'leiden' or 'louvain'.",
    )
    use_rna_velocity: bool = Field(
        default=False,
        description="Use RNA velocity to orient edges in the abstracted graph and estimate transitions. Requires that adata.uns contains a directed single-cell graph with key ['velocity_graph'].",
    )
    model: Literal["v1.2", "v1.0"] = Field(
        default="v1.2", description="The PAGA connectivity model."
    )
    neighbors_key: Optional[str] = Field(
        default=None,
        description="If specified, paga looks .uns[neighbors_key] for neighbors settings and uses the corresponding connectivities and distances.",
    )

    @field_validator("model")
    def validate_Param(cls, v: str) -> str:
        """Validate model version is supported"""
        if v not in ["v1.2", "v1.0"]:
            raise ValueError("model must be either 'v1.2' or 'v1.0'")
        return v


class IngestParam(BaseModel):
    """Input schema for the ingest tool that maps labels and embeddings from reference data to new data."""

    adata: str = Field(..., description="The AnnData object variable name.")
    adata_ref: str = Field(
        ..., description="The reference AnnData object variable name."
    )
    obs: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="Labels' keys in adata_ref.obs which need to be mapped to adata.obs (inferred for observation of adata).",
    )
    embedding_method: Union[str, List[str]] = Field(
        default=["umap", "pca"],
        description="Embeddings in adata_ref which need to be mapped to adata. The only supported values are 'umap' and 'pca'.",
    )
    labeling_method: str = Field(
        default="knn",
        description="The method to map labels in adata_ref.obs to adata.obs. The only supported value is 'knn'.",
    )
    neighbors_key: Optional[str] = Field(
        default=None,
        description="If specified, ingest looks adata_ref.uns[neighbors_key] for neighbors settings and uses the corresponding distances.",
    )

    @field_validator("embedding_method")
    def validate_embedding_method(
        cls, v: Union[str, List[str]]
    ) -> Union[str, List[str]]:
        """Validate embedding method is supported"""
        valid_methods = ["umap", "pca"]

        if isinstance(v, str):
            if v.lower() not in valid_methods:
                raise ValueError(f"embedding_method must be one of {valid_methods}")
            return v.lower()

        elif isinstance(v, list):
            for method in v:
                if method.lower() not in valid_methods:
                    raise ValueError(
                        f"embedding_method must contain only values from {valid_methods}"
                    )
            return [method.lower() for method in v]

        return v

    @field_validator("labeling_method")
    def validate_labeling_method(cls, v: str) -> str:
        """Validate labeling method is supported"""
        if v.lower() != "knn":
            raise ValueError("labeling_method must be 'knn'")
        return v.lower()


class RankGenesGroupsParam(BaseModel):
    """Input schema for the rank_genes_groups tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    groupby: str = Field(
        ...,  # Required field
        description="The key of the observations grouping to consider.",
    )
    mask_var: Optional[Union[str, List[bool]]] = Field(
        default=None, description="Select subset of genes to use in statistical tests."
    )
    use_raw: Optional[bool] = Field(
        default=None, description="Use raw attribute of adata if present."
    )
    groups: Union[Literal["all"], List[str]] = Field(
        default="all",
        description="Subset of groups to which comparison shall be restricted, or 'all' for all groups.",
    )
    reference: str = Field(
        default="rest",
        description="If 'rest', compare each group to the union of the rest of the group. If a group identifier, compare with respect to this group.",
    )
    n_genes: Optional[int] = Field(
        default=None,
        description="The number of genes that appear in the returned tables. Defaults to all genes.",
        gt=0,
    )
    rankby_abs: bool = Field(
        default=False,
        description="Rank genes by the absolute value of the score, not by the score.",
    )
    pts: bool = Field(
        default=False, description="Compute the fraction of cells expressing the genes."
    )
    key_added: Optional[str] = Field(
        default=None, description="The key in adata.uns information is saved to."
    )
    method: Optional[str] = Field(
        default=None,
        description="Method for differential expression analysis. Default is 't-test'.",
    )
    corr_method: str = Field(
        default="benjamini-hochberg",
        description="p-value correction method. Used only for 't-test', 't-test_overestim_var', and 'wilcoxon'.",
    )
    tie_correct: bool = Field(
        default=False,
        description="Use tie correction for 'wilcoxon' scores. Used only for 'wilcoxon'.",
    )
    layer: Optional[str] = Field(
        default=None,
        description="Key from adata.layers whose value will be used to perform tests on.",
    )

    @field_validator("method")
    def validate_method(cls, v: Optional[str]) -> Optional[str]:
        """Validate method is supported"""
        if v is not None:
            valid_methods = ["t-test", "t-test_overestim_var", "wilcoxon", "logreg"]
            if v not in valid_methods:
                raise ValueError(f"method must be one of {valid_methods}")
        return v

    @field_validator("corr_method")
    def validate_corr_method(cls, v: str) -> str:
        """Validate correction method is supported"""
        valid_methods = ["benjamini-hochberg", "bonferroni"]
        if v not in valid_methods:
            raise ValueError(f"corr_method must be one of {valid_methods}")
        return v

    @field_validator("n_genes")
    def validate_n_genes(cls, v: Optional[int]) -> Optional[int]:
        """Validate n_genes is positive"""
        if v is not None and v <= 0:
            raise ValueError("n_genes must be a positive integer")
        return v


class FilterRankGenesGroupsParam(BaseModel):
    """Input schema for filtering ranked genes groups."""

    adata: str = Field(..., description="The AnnData object variable name.")
    key: Optional[str] = Field(
        default=None,
        description="Key from adata.uns where rank_genes_groups output is stored.",
    )
    groupby: Optional[str] = Field(
        default=None, description="The key of the observations grouping to consider."
    )
    use_raw: Optional[bool] = Field(
        default=None, description="Use raw attribute of adata if present."
    )
    key_added: str = Field(
        default="rank_genes_groups_filtered",
        description="The key in adata.uns information is saved to.",
    )
    min_in_group_fraction: float = Field(
        default=0.25,
        description="Minimum fraction of cells expressing the gene within the group.",
        ge=0.0,
        le=1.0,
    )
    min_fold_change: Union[int, float] = Field(
        default=1,
        description="Minimum fold change for a gene to be considered significant.",
        gt=0,
    )
    max_out_group_fraction: float = Field(
        default=0.5,
        description="Maximum fraction of cells expressing the gene outside the group.",
        ge=0.0,
        le=1.0,
    )
    compare_abs: bool = Field(
        default=False,
        description="If True, compare absolute values of log fold change with min_fold_change.",
    )

    @field_validator("min_in_group_fraction", "max_out_group_fraction")
    def validate_fractions(cls, v: float) -> float:
        """Validate fractions are between 0 and 1"""
        if v < 0 or v > 1:
            raise ValueError("Fraction values must be between 0 and 1")
        return v

    @field_validator("min_fold_change")
    def validate_fold_change(cls, v: Union[int, float]) -> Union[int, float]:
        """Validate min_fold_change is positive"""
        if v <= 0:
            raise ValueError("min_fold_change must be a positive number")
        return v


class MarkerGeneOverlapParam(BaseModel):
    """Input schema for the marker gene overlap tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    adata_ref: str = Field(
        ..., description="The reference AnnData object variable name."
    )
    key: str = Field(
        default="rank_genes_groups",
        description="The key in adata.uns where the rank_genes_groups output is stored.",
    )
    method: str = Field(
        default="overlap_count",
        description="Method to calculate marker gene overlap: 'overlap_count', 'overlap_coef', or 'jaccard'.",
    )
    normalize: Optional[Literal["reference", "data"]] = Field(
        default=None,
        description="Normalization option for the marker gene overlap output. Only applicable when method is 'overlap_count'.",
    )
    top_n_markers: Optional[int] = Field(
        default=None,
        description="The number of top data-derived marker genes to use. By default the top 100 marker genes are used.",
        gt=0,
    )
    adj_pval_threshold: Optional[float] = Field(
        default=None,
        description="A significance threshold on the adjusted p-values to select marker genes.",
        gt=0,
        le=1.0,
    )
    key_added: str = Field(
        default="marker_gene_overlap",
        description="Name of the .uns field that will contain the marker overlap scores.",
    )

    @field_validator("method")
    def validate_method(cls, v: str) -> str:
        """Validate method is supported"""
        valid_methods = ["overlap_count", "overlap_coef", "jaccard"]
        if v not in valid_methods:
            raise ValueError(f"method must be one of {valid_methods}")
        return v

    @field_validator("normalize")
    def validate_normalize(
        cls, v: Optional[str], info: ValidationInfo
    ) -> Optional[str]:
        """Validate normalize is only used with overlap_count method"""
        if v is not None:
            if v not in ["reference", "data"]:
                raise ValueError("normalize must be either 'reference' or 'data'")

            values = info.data
            if "method" in values and values["method"] != "overlap_count":
                raise ValueError(
                    "normalize can only be used when method is 'overlap_count'"
                )
        return v

    @field_validator("top_n_markers")
    def validate_top_n_markers(cls, v: Optional[int]) -> Optional[int]:
        """Validate top_n_markers is positive"""
        if v is not None and v <= 0:
            raise ValueError("top_n_markers must be a positive integer")
        return v

    @field_validator("adj_pval_threshold")
    def validate_adj_pval_threshold(cls, v: Optional[float]) -> Optional[float]:
        """Validate adj_pval_threshold is between 0 and 1"""
        if v is not None and (v <= 0 or v > 1):
            raise ValueError("adj_pval_threshold must be between 0 and 1")
        return v


class ScoreGenesParam(BaseModel):
    """Input schema for the score_genes tool that calculates gene scores based on average expression."""

    adata: str = Field(..., description="The AnnData object variable name.")
    gene_list: List[str] = Field(..., description="List of genes to score.")
    ctrl_size: int = Field(
        default=50,
        description="Number of reference genes to be sampled from each bin.",
        gt=0,
    )
    gene_pool: Optional[List[str]] = Field(
        default=None,
        description="Genes for sampling the reference set. Default is all genes.",
    )
    n_bins: int = Field(
        default=25, description="Number of expression level bins for sampling.", gt=0
    )
    score_name: str = Field(
        default="score", description="Name of the field to be added in .obs."
    )
    random_state: int = Field(default=0, description="The random seed for sampling.")
    use_raw: Optional[bool] = Field(
        default=None,
        description="Whether to use raw attribute of adata. Defaults to True if .raw is present.",
    )

    @field_validator("ctrl_size", "n_bins")
    def validate_positive_integers(cls, v: int) -> int:
        """Validate positive integers"""
        if v <= 0:
            raise ValueError("must be a positive integer")
        return v


class ScoreGenesCellCycleParam(BaseModel):
    """Input schema for the score_genes_cell_cycle tool that scores cell cycle genes."""

    adata: str = Field(..., description="The AnnData object variable name.")
    s_genes: List[str] = Field(
        ...,  # Required field
        description="List of genes associated with S phase.",
    )
    g2m_genes: List[str] = Field(
        ...,  # Required field
        description="List of genes associated with G2M phase.",
    )
    gene_pool: Optional[List[str]] = Field(
        default=None,
        description="Genes for sampling the reference set. Default is all genes.",
    )
    n_bins: int = Field(
        default=25, description="Number of expression level bins for sampling.", gt=0
    )
    score_name: Optional[str] = Field(
        default=None,
        description="Name of the field to be added in .obs. If None, the scores are added as 'S_score' and 'G2M_score'.",
    )
    random_state: int = Field(default=0, description="The random seed for sampling.")
    use_raw: Optional[bool] = Field(
        default=None,
        description="Whether to use raw attribute of adata. Defaults to True if .raw is present.",
    )

    @field_validator("s_genes", "g2m_genes")
    def validate_gene_lists(cls, v: List[str]) -> List[str]:
        """Validate gene lists are not empty"""
        if len(v) == 0:
            raise ValueError("Gene list cannot be empty")
        return v

    @field_validator("n_bins")
    def validate_positive_integers(cls, v: int) -> int:
        """Validate positive integers"""
        if v <= 0:
            raise ValueError("n_bins must be a positive integer")
        return v


class PCAParam(BaseModel):
    """Input schema for the PCA preprocessing tool."""

    adata: str = Field(..., description="The AnnData object variable name.")
    n_comps: Optional[int] = Field(
        default=None,
        description="Number of principal components to compute. Defaults to 50 or 1 - minimum dimension size.",
        gt=0,
    )

    layer: Optional[str] = Field(
        default=None, description="If provided, which element of layers to use for PCA."
    )

    zero_center: Optional[bool] = Field(
        default=True,
        description="If True, compute standard PCA from covariance matrix.",
    )

    svd_solver: Optional[Literal["arpack", "randomized", "auto", "lobpcg", "tsqr"]] = (
        Field(default=None, description="SVD solver to use.")
    )
    mask_var: Optional[Union[str, bool]] = Field(
        default=None,
        description="Boolean mask or string referring to var column for subsetting genes.",
    )
    dtype: str = Field(
        default="float32", description="Numpy data type string for the result."
    )
    chunked: bool = Field(
        default=False, description="If True, perform an incremental PCA on segments."
    )

    chunk_size: Optional[int] = Field(
        default=None,
        description="Number of observations to include in each chunk.",
        gt=0,
    )

    key_added: str = Field(
        default="X_pca", description="PCA embedding stored key in adata.obsm."
    )

    @field_validator("n_comps", "chunk_size")
    def validate_positive_integers(cls, v: Optional[int]) -> Optional[int]:
        """Validate positive integers"""
        if v is not None and v <= 0:
            raise ValueError("must be a positive integer")
        return v



import argparse
from typing import Optional, Dict, Callable
from .util import add_figure_route, set_env
import os


class MCPCLI:
    """Base class for CLI applications with support for dynamic modules and parameters."""

    def __init__(self, name: str, help_text: str, mcp=None, manager=None):
        self.name = name
        self.mcp = mcp
        self.manager = manager
        self.parser = argparse.ArgumentParser(description=help_text, prog=name)
        self.subcommands: Dict[str, tuple[argparse.ArgumentParser, Callable]] = {}
        self._setup_commands()

    def _setup_commands(self):
        """Setup the main commands for the CLI."""
        subparsers = self.parser.add_subparsers(
            dest="command", help="Available commands"
        )
        run_parser = subparsers.add_parser(
            "run", help="Start the server with the specified configuration"
        )
        self._setup_run_command(run_parser)
        self.subcommands["run"] = (run_parser, self._run_command)

    def _setup_run_command(self, parser: argparse.ArgumentParser):
        """Setup run command arguments."""
        parser.add_argument(
            "-t",
            "--transport",
            default="stdio",
            choices=["stdio", "shttp", "sse"],
            help="specify transport type",
        )
        parser.add_argument(
            "-p", "--port", type=int, default=8000, help="transport port"
        )
        parser.add_argument("--host", default="127.0.0.1", help="transport host")
        parser.add_argument("-f", "--forward", help="forward request to another server")
        parser.add_argument(
            "-wd", "--working-dir", default=".", help="working directory"
        )
        parser.add_argument(
            "--run-mode", choices=["tool", "code"], default="code", help="run mode"
        )
        parser.add_argument(
            "--tool-mode",
            choices=["auto", "normal"],
            default="normal",
            help="tool selection mode",
        )
        parser.add_argument("--log-file", help="log file path, use stdout if None")

    def add_command(
        self, name: str, help_text: str, handler: Callable
    ) -> argparse.ArgumentParser:
        """add new subcommand

        Args:
            name: subcommand name
            help_text: help text
            handler: handler function

        Returns:
            ArgumentParser: parser for the subcommand
        """
        subparsers = self.parser._subparsers._group_actions[0]
        parser = subparsers.add_parser(name, help=help_text)
        self.subcommands[name] = (parser, handler)
        return parser

    def get_command_parser(self, name: str) -> Optional[argparse.ArgumentParser]:
        """get the parser for the subcommand

        Args:
            name: subcommand name

        Returns:
            ArgumentParser: parser for the subcommand, return None if the subcommand does not exist
        """
        if name in self.subcommands:
            return self.subcommands[name][0]
        return None

    def _run_command(self, args):
        """Start the server with the specified configuration."""
        os.chdir(args.working_dir)
        if hasattr(args, "module"):
            if "all" in args.module:
                modules = None
            elif isinstance(args.module, list) and bool(args.module):
                modules = args.module
        else:
            modules = None
        if self.manager is not None:
            from .backend import NotebookManager, AdataManager

            if args.run_mode == "code":
                backend = NotebookManager
                self.mcp = self.manager(
                    self.name, include_tags=["nb", "rag"], backend=backend
                ).mcp
            else:
                backend = AdataManager
                self.mcp = self.manager(
                    self.name, include_modules=modules, backend=backend
                ).mcp
                all_tools = self.mcp._tool_manager._tools
                auto_tools = {
                    tool: all_tools[tool]
                    for tool in all_tools
                    if "auto" in all_tools[tool].tags
                }
                if args.tool_mode == "auto":
                    all_tools = self.mcp._tool_manager._tools
                    self.mcp._tool_manager._all_tools = all_tools
                    self.mcp._tool_manager._tools = auto_tools
                else:
                    for name in auto_tools:
                        self.mcp._tool_manager.remove_tool(name)
        elif self.mcp is not None:
            pass
        else:
            raise ValueError("No manager or mcp provided")
        transport = args.transport
        self.run_mcp(args.log_file, args.forward, transport, args.host, args.port)

    def run_mcp(self, log_file, forward, transport, host, port):
        set_env(log_file, forward, transport, host, port)
        from .logging_config import setup_logger

        setup_logger(log_file)
        if transport == "stdio":
            self.mcp.run()
        elif transport in ["sse", "shttp"]:
            transport = "streamable-http" if transport == "shttp" else transport
            add_figure_route(self.mcp)
            self.mcp.run(transport=transport, host=host, port=port, log_level="info")

    def run(self):
        """Run the CLI application."""
        args = self.parser.parse_args()
        if args.command in self.subcommands:
            handler = self.subcommands[args.command][1]
            handler(args)
        else:
            self.parser.print_help()



from .io import ScanpyIOMCP, io_mcp
from .pl import ScanpyPlottingMCP
from .pp import ScanpyPreprocessingMCP
from .tl import ScanpyToolsMCP
from .util import ScanpyUtilMCP

__all__ = [
    "io_mcp",
    "ScanpyIOMCP",
    "ScanpyPlottingMCP",
    "ScanpyPreprocessingMCP",
    "ScanpyToolsMCP",
    "ScanpyUtilMCP",
]



from pydantic import Field, field_validator, model_validator, BaseModel
from typing import Optional, Literal


class ReadParam(BaseModel):
    """Input schema for the read tool."""

    filename: str = Field(..., description="Path to the file to read.")

    backed: Literal["r", "r+"] = Field(
        default=None,
        description="If 'r', load AnnData in 'backed' mode instead of fully loading it into memory ('memory' mode). If you want to modify backed attributes of the AnnData object, you need to choose 'r+'.",
    )
    sheet: str = Field(
        default=None, description="Name of sheet/table in hdf5 or Excel file."
    )
    delimiter: str = Field(
        default=None,
        description="Delimiter that separates data within text file. If None, will split at arbitrary number of white spaces, which is different from enforcing splitting at any single white space.",
    )
    first_column_names: bool = Field(
        default=False,
        description="Assume the first column stores row names. This is only necessary if these are not strings: strings in the first column are automatically assumed to be row names.",
    )
    transpose: bool = Field(
        default=False, description="If True, the data will be transposed."
    )
    backup_url: str = Field(
        default=None,
        description="Retrieve the file from an URL if not present on disk.",
    )
    cache: bool = Field(
        default=False,
        description="If False, read from source, if True, read from fast 'h5ad' cache.",
    )
    cache_compression: Literal["gzip", "lzf"] = Field(
        default=None,
        description="See the h5py dataset_compression. (Default: settings.cache_compression)",
    )
    var_names: str = Field(
        default="gene_symbols",
        description="The variables index for 10x mtx format. Either 'gene_symbols' or 'gene_ids'.",
    )
    make_unique: bool = Field(
        default=True,
        description="Whether to make the variables index unique by appending '-1', '-2' etc. or not. Used for 10x mtx format.",
    )
    gex_only: bool = Field(
        default=True,
        description="Only keep 'Gene Expression' data and ignore other feature types, e.g. 'Antibody Capture', 'CRISPR Guide Capture', or 'Custom'. Used for 10x formats.",
    )
    prefix: str = Field(
        default=None,
        description="Any prefix before matrix.mtx, genes.tsv and barcodes.tsv. For instance, if the files are named patientA_matrix.mtx, patientA_genes.tsv and patientA_barcodes.tsv the prefix is patientA_. Used for 10x mtx format.",
    )

    @field_validator("backed")
    def validate_backed(cls, v: Optional[str]) -> Optional[str]:
        if v is not None and v not in ["r", "r+"]:
            raise ValueError("If backed is provided, it must be either 'r' or 'r+'")
        return v

    @field_validator("cache_compression")
    def validate_cache_compression(cls, v: Optional[str]) -> Optional[str]:
        if v is not None and v not in ["gzip", "lzf"]:
            raise ValueError("cache_compression must be either 'gzip', 'lzf', or None")
        return v

    @field_validator("var_names")
    def validate_var_names(cls, v: Optional[str]) -> Optional[str]:
        if v is not None and v not in ["gene_symbols", "gene_ids"]:
            raise ValueError("var_names must be either 'gene_symbols' or 'gene_ids'")
        return v


class WriteParam(BaseModel):
    """Input schema for the write tool."""

    filename: str = Field(
        description="Path to save the file. If no extension is provided, the default format will be used."
    )
    ext: Literal["h5", "csv", "txt", "npz"] = Field(
        default=None,
        description="File extension to infer file format. If None, defaults to scanpy's settings.file_format_data.",
    )
    compression: Literal["gzip", "lzf"] = Field(
        default="gzip", description="Compression format for h5 files."
    )
    compression_opts: int = Field(
        default=None, description="Compression options for h5 files."
    )

    @field_validator("filename")
    def validate_filename(cls, v: str) -> str:
        # Allow any filename since the extension is optional and can be inferred
        return v

    @model_validator(mode="after")
    def validate_extension_compression(self) -> "WriteParam":
        # If ext is provided and not h5, compression should be None
        if self.ext is not None and self.ext != "h5" and self.compression is not None:
            raise ValueError("Compression can only be used with h5 files")
        return self



import pytest
from fastmcp import Client
import nest_asyncio

# Apply nest_asyncio at module level
nest_asyncio.apply()


@pytest.mark.asyncio
async def test_tool_selection(mcp):
    task = """
    read the file  /date/test.h5ad , filter the cells with ncells>100
    """
    async with Client(mcp) as client:
        result = await client.call_tool("auto_search_tool", {"task": task})
        assert "io_read" in result.content[0].text



from pydantic import Field, BaseModel
from typing import Optional, List, Dict, Any, Literal


class MarkVarParam(BaseModel):
    """Determine or mark if each gene meets specific conditions and store results in adata.var as boolean values"""

    var_name: str = Field(
        default=None,
        description="Column name that will be added to adata.var, do not set if user does not ask",
    )
    pattern_type: Optional[Literal["startswith", "endswith", "contains"]] = Field(
        default=None,
        description="Pattern matching type (startswith/endswith/contains), it should be None when gene_class is not None",
    )
    patterns: Optional[str] = Field(
        default=None,
        description="gene pattern to match, must be a string, it should be None when gene_class is not None",
    )

    gene_class: Optional[Literal["mitochondrion", "ribosomal", "hemoglobin"]] = Field(
        default=None, description="Gene class type (Mitochondrion/Ribosomal/Hemoglobin)"
    )


class ListVarParam(BaseModel):
    """ListVarModel"""

    pass


class ListObsParam(BaseModel):
    """ListObsModel"""

    pass


class VarNamesParam(BaseModel):
    """ListObsModel"""

    var_names: List[str] = Field(default=None, description="gene names.")


class ConcatBaseParam(BaseModel):
    """Model for concatenating AnnData objects"""

    axis: Literal["obs", 0, "var", 1] = Field(
        default="obs",
        description="Which axis to concatenate along. 'obs' or 0 for observations, 'var' or 1 for variables.",
    )
    join: Literal["inner", "outer"] = Field(
        default="inner",
        description="How to align values when concatenating. If 'outer', the union of the other axis is taken. If 'inner', the intersection.",
    )
    merge: Optional[Literal["same", "unique", "first", "only"]] = Field(
        default=None,
        description="How elements not aligned to the axis being concatenated along are selected.",
    )
    uns_merge: Optional[Literal["same", "unique", "first", "only"]] = Field(
        default=None,
        description="How the elements of .uns are selected. Uses the same set of strategies as the merge argument, except applied recursively.",
    )
    label: Optional[str] = Field(
        default=None,
        description="label different adata, Column in axis annotation (i.e. .obs or .var) to place batch information in. ",
    )
    keys: Optional[List[str]] = Field(
        default=None,
        description="Names for each object being added. These values are used for column values for label or appended to the index if index_unique is not None.",
    )
    index_unique: Optional[str] = Field(
        default=None,
        description="Whether to make the index unique by using the keys. If provided, this is the delimiter between '{orig_idx}{index_unique}{key}'.",
    )
    fill_value: Optional[Any] = Field(
        default=None,
        description="When join='outer', this is the value that will be used to fill the introduced indices.",
    )
    pairwise: bool = Field(
        default=False,
        description="Whether pairwise elements along the concatenated dimension should be included.",
    )


class DPTIROOTParam(BaseModel):
    """Input schema for setting the root cell for diffusion pseudotime."""

    diffmap_key: str = Field(
        default="X_diffmap",
        description="Key for diffusion map coordinates stored in adata.obsm.",
    )
    dimension: int = Field(
        description="Dimension index to use for finding the root cell."
    )
    direction: Literal["min", "max"] = Field(
        description="use the minimum or maximum value along the selected dimension to identify the root cell."
    )


class CelltypeMapCellTypeParam(BaseModel):
    """Input schema for mapping cluster IDs to cell type names."""

    cluster_key: str = Field(description="Key in adata.obs containing cluster IDs.")
    added_key: str = Field(description="Key to add to adata.obs for cell type names.")
    mapping: Dict[str, str] = Field(
        default=None,
        description="Mapping Dictionary from cluster IDs to cell type names.",
    )
    new_names: Optional[List[str]] = Field(
        default=None, description="a list of new cell type names."
    )


class AddLayerParam(BaseModel):
    """Input schema for adding a layer to AnnData object."""

    layer_name: str = Field(description="Name of the layer to add to adata.layers.")


class QueryOpLogParam(BaseModel):
    """QueryOpLogModel"""

    n: int = Field(default=10, description="Number of operations to return.")



import scanpy as sc
from fastmcp.exceptions import ToolError
from fastmcp.tools.tool import Tool
from scmcp_shared.schema.preset.tl import *
from scmcp_shared.schema.preset import AdataInfo
from scmcp_shared.util import (
    filter_args,
    add_op_log,
    forward_request,
    get_ads,
    generate_msg,
)
from scmcp_shared.mcp_base import BaseMCP


class ScanpyToolsMCP(BaseMCP):
    def __init__(
        self,
        include_tools: list = None,
        exclude_tools: list = None,
        AdataInfo: AdataInfo = AdataInfo,
    ):
        """
        Initialize ScanpyMCP with optional tool filtering.

        Args:
            include_tools (list, optional): List of tool names to include. If None, all tools are included.
            exclude_tools (list, optional): List of tool names to exclude. If None, no tools are excluded.
            AdataInfo: The AdataInfo class to use for type annotations.
        """
        super().__init__("ScanpyMCP-TL-Server", include_tools, exclude_tools, AdataInfo)

    def _tool_tsne(self):
        def _tsne(
            request: TSNEParam = TSNEParam(), adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """t-distributed stochastic neighborhood embedding (t-SNE) for visualization"""
            try:
                result = forward_request("tl_tsne", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.tsne)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.tsne(adata, **func_kwargs)
                add_op_log(adata, sc.tl.tsne, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_tsne, name="tsne", enabled=True)

    def _tool_umap(self):
        def _umap(
            request: UMAPParam = UMAPParam(), adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Uniform Manifold Approximation and Projection (UMAP) for visualization"""
            try:
                result = forward_request("tl_umap", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.umap)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.umap(adata, **func_kwargs)
                add_op_log(adata, sc.tl.umap, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_umap, name="umap", enabled=True)

    def _tool_draw_graph(self):
        def _draw_graph(
            request: DrawGraphParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Force-directed graph drawing"""
            try:
                result = forward_request("tl_draw_graph", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.draw_graph)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.draw_graph(adata, **func_kwargs)
                add_op_log(adata, sc.tl.draw_graph, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_draw_graph, name="draw_graph", enabled=True)

    def _tool_diffmap(self):
        def _diffmap(request: DiffMapParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """Diffusion Maps for dimensionality reduction"""
            try:
                result = forward_request("tl_diffmap", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.diffmap)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.diffmap(adata, **func_kwargs)
                adata.obsm["X_diffmap"] = adata.obsm["X_diffmap"][:, 1:]
                add_op_log(adata, sc.tl.diffmap, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_diffmap, name="diffmap", enabled=True)

    def _tool_embedding_density(self):
        def _embedding_density(
            request: EmbeddingDensityParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Calculate the density of cells in an embedding"""
            try:
                result = forward_request("tl_embedding_density", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.embedding_density)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.embedding_density(adata, **func_kwargs)
                add_op_log(adata, sc.tl.embedding_density, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _embedding_density, name="embedding_density", enabled=True
        )

    def _tool_leiden(self):
        def _leiden(
            request: LeidenParam = LeidenParam(),
            adinfo: self.AdataInfo = self.AdataInfo(),
        ):
            """Leiden clustering algorithm for community detection"""
            try:
                result = forward_request("tl_leiden", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.leiden)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.leiden(adata, **func_kwargs)
                add_op_log(adata, sc.tl.leiden, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_leiden, name="leiden", enabled=True)

    def _tool_louvain(self):
        def _louvain(
            request: LouvainParam = LouvainParam(),
            adinfo: self.AdataInfo = self.AdataInfo(),
        ):
            """Louvain clustering algorithm for community detection"""
            try:
                result = forward_request("tl_louvain", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.louvain)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.louvain(adata, **func_kwargs)
                add_op_log(adata, sc.tl.louvain, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_louvain, name="louvain", enabled=True)

    def _tool_dendrogram(self):
        def _dendrogram(
            request: DendrogramParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Hierarchical clustering dendrogram"""
            try:
                result = forward_request("tl_dendrogram", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.dendrogram)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.dendrogram(adata, **func_kwargs)
                add_op_log(adata, sc.tl.dendrogram, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_dendrogram, name="dendrogram", enabled=True)

    def _tool_dpt(self):
        def _dpt(request: DPTParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """Diffusion Pseudotime (DPT) analysis"""
            try:
                result = forward_request("tl_dpt", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.dpt)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.dpt(adata, **func_kwargs)
                add_op_log(adata, sc.tl.dpt, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_dpt, name="dpt", enabled=True)

    def _tool_paga(self):
        def _paga(request: PAGAParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """Partition-based graph abstraction"""
            try:
                result = forward_request("tl_paga", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.paga)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.paga(adata, **func_kwargs)
                add_op_log(adata, sc.tl.paga, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_paga, name="paga", enabled=True)

    def _tool_ingest(self):
        def _ingest(request: IngestParam, adinfo: self.AdataInfo = self.AdataInfo()):
            """Map labels and embeddings from reference data to new data"""
            try:
                result = forward_request("tl_ingest", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.ingest)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.ingest(adata, **func_kwargs)
                add_op_log(adata, sc.tl.ingest, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_ingest, name="ingest", enabled=True)

    def _tool_rank_genes_groups(self):
        def _rank_genes_groups(
            request: RankGenesGroupsParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Rank genes for characterizing groups, for differentially expressison analysis"""
            try:
                result = forward_request("tl_rank_genes_groups", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.rank_genes_groups)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.rank_genes_groups(adata, **func_kwargs)
                add_op_log(adata, sc.tl.rank_genes_groups, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _rank_genes_groups, name="rank_genes_groups", enabled=True
        )

    def _tool_filter_rank_genes_groups(self):
        def _filter_rank_genes_groups(
            request: FilterRankGenesGroupsParam,
            adinfo: self.AdataInfo = self.AdataInfo(),
        ):
            """Filter out genes based on fold change and fraction of genes"""
            try:
                result = forward_request("tl_filter_rank_genes_groups", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.filter_rank_genes_groups)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.filter_rank_genes_groups(adata, **func_kwargs)
                add_op_log(adata, sc.tl.filter_rank_genes_groups, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _filter_rank_genes_groups, name="filter_rank_genes_groups", enabled=True
        )

    def _tool_marker_gene_overlap(self):
        def _marker_gene_overlap(
            request: MarkerGeneOverlapParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Calculate overlap between data-derived marker genes and reference markers"""
            try:
                result = forward_request("tl_marker_gene_overlap", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.marker_gene_overlap)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.marker_gene_overlap(adata, **func_kwargs)
                add_op_log(adata, sc.tl.marker_gene_overlap, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _marker_gene_overlap, name="marker_gene_overlap", enabled=True
        )

    def _tool_score_genes(self):
        def _score_genes(
            request: ScoreGenesParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Score a set of genes based on their average expression"""
            try:
                result = forward_request("tl_score_genes", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.score_genes)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.score_genes(adata, **func_kwargs)
                add_op_log(adata, sc.tl.score_genes, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_score_genes, name="score_genes", enabled=True)

    def _tool_score_genes_cell_cycle(self):
        def _score_genes_cell_cycle(
            request: ScoreGenesCellCycleParam, adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Score cell cycle genes and assign cell cycle phases"""
            try:
                result = forward_request("tl_score_genes_cell_cycle", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.score_genes_cell_cycle)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.score_genes_cell_cycle(adata, **func_kwargs)
                add_op_log(adata, sc.tl.score_genes_cell_cycle, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(
            _score_genes_cell_cycle, name="score_genes_cell_cycle", enabled=True
        )

    def _tool_pca(self):
        def _pca(
            request: PCAParam = PCAParam(), adinfo: self.AdataInfo = self.AdataInfo()
        ):
            """Compute PCA (Principal Component Analysis)."""
            try:
                result = forward_request("tl_pca", request, adinfo)
                if result is not None:
                    return result
                func_kwargs = filter_args(request, sc.tl.pca)
                ads = get_ads()
                adata = ads.get_adata(adinfo=adinfo)
                sc.tl.pca(adata, **func_kwargs)
                add_op_log(adata, sc.tl.pca, func_kwargs, adinfo)
                return [generate_msg(adinfo, adata, ads)]
            except ToolError as e:
                raise ToolError(e)
            except Exception as e:
                if hasattr(e, "__context__") and e.__context__:
                    raise ToolError(e.__context__)
                else:
                    raise ToolError(e)

        return Tool.from_function(_pca, name="pca", enabled=True)
